import 'dart:async';

import 'package:adeptannotations/adeptannotations.dart';
import 'package:adeptgenerator/src/base_generator_for_annotation.dart';
import 'package:adeptgenerator/src/config/copy_with_key_config.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

class CopyWithGenerator
    extends BaseGeneratorForAnnotation<CopyWith, CopyWithKey> {
  final CopyWith globalAnnotation;
  final bool defaultIgnoreSynthetic = false;
  final bool defaultAllowNullable = false;
  final bool defaultConsiderEmptyCollectionAsNull = true;
  final bool defaultIncludeIfPrivate = false;
  final bool defaultAvoidConstructor = false;

  CopyWithGenerator(BuilderOptions options)
      : globalAnnotation = CopyWith(
          ignoreSynthetic: options.config['copy_with_ignore_synthetic'],
          allowNullable: options.config['copy_with_allow_nullable'],
          considerEmptyCollectionAsNull:
              options.config['copy_with_consider_empty_collection_as_null'],
          includePrivateFields:
              options.config['copy_with_include_private_fields'],
          avoidConstructor: options.config['copy_with_avoid_constructor'],
        );

  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement =
        getClassElement(element, annotationName: '@CopyWith()');
    String copyWithFunction =
        _generateCopyWithFunction(classElement, annotation);
    return '''
    $copyWithFunction
    ''';
  }

  /// This function manages the logic of generating the copyWith boilerplate code
  String _generateCopyWithFunction(
      ClassElement classElement, ConstantReader annotation) {
    // checkDefaultConstructor(classElement, annotationName: '@CopyWith()');

    String classType = classElement.displayName;

    Set<String> copyWithFunctionParamsSet = {};
    Set<String> copyWithConstructorLogicSet = {};
    Set<String> copyWithExtraLogicSet = {};
    final copyWithAnnotation = _getCopyWith(classElement) ?? const CopyWith();

    void fieldAnalyzer(FieldElement field) {
      final copyWithKey = _getCopyWithKey(field) ?? const CopyWithKeyConfig();
      final ignoreSynthetic = copyWithKey.ignoreSynthetic ??
          copyWithAnnotation.ignoreSynthetic ??
          globalAnnotation.ignoreSynthetic ??
          defaultIgnoreSynthetic;
      final allowNullable = copyWithKey.allowNullable ??
          copyWithAnnotation.allowNullable ??
          globalAnnotation.allowNullable ??
          defaultAllowNullable;
      final considerEmptyCollectionAsNull =
          copyWithKey.considerEmptyCollectionAsNull ??
              copyWithAnnotation.considerEmptyCollectionAsNull ??
              globalAnnotation.considerEmptyCollectionAsNull ??
              defaultConsiderEmptyCollectionAsNull;
      final includeIfPrivate = copyWithKey.includeIfPrivate ??
          copyWithAnnotation.includePrivateFields ??
          globalAnnotation.includePrivateFields ??
          defaultIncludeIfPrivate;
      final avoidConstructor = copyWithKey.avoidConstructor ??
          copyWithAnnotation.avoidConstructor ??
          globalAnnotation.avoidConstructor ??
          defaultAvoidConstructor;

      if (field.isFinal && (field.isPrivate || avoidConstructor)) return;
      if (field.isSynthetic && ignoreSynthetic) return;
      if (field.isConst || field.isStatic) return;
      if (field.isSynthetic && (field.getter == null || field.setter == null))
        return;
      if (field.isPrivate && !includeIfPrivate) return;
      if (copyWithKey.ignore) return;

      final isNullable = field.type.nullabilitySuffix != NullabilitySuffix.none;
      final isCollection = field.type.isDartCoreList ||
          field.type.isDartCoreSet ||
          field.type.isDartCoreMap ||
          field.type.isDartCoreIterable;
      String fieldType = field.type.getDisplayString(withNullability: false);
      String fieldName = field.displayName;
      String currentData = field.isPrivate ? fieldName : 'this.$fieldName';
      String newData = fieldName.substring(field.isPrivate ? 1 : 0);
      if (considerEmptyCollectionAsNull && isCollection) {
        newData = '(($fieldName?.isNotEmpty ?? false) ? $fieldName : null)';
      }
      String copyWithDataFieldValue =
          allowNullable ? newData : '$newData ?? $currentData';

      //Default value applies only if field is nullable, otherwise is not necessary because of null-safety
      bool defaultValueApplies = isNullable;
      if (copyWithKey.copyWithFunction != null) {
        copyWithDataFieldValue = copyWithKey.copyWithFunction!;
      }

      if (copyWithKey.defaultValue != null && defaultValueApplies) {
        copyWithDataFieldValue =
            '($copyWithDataFieldValue) ?? ${copyWithKey.defaultValue}';
      }
      copyWithFunctionParamsSet.add(
          '$fieldType${fieldType == 'dynamic' ? '' : '?'} ${fieldName.substring(field.isPrivate ? 1 : 0)},');
      if (field.isSynthetic || field.isPrivate || avoidConstructor) {
        copyWithExtraLogicSet.add('..$fieldName = $copyWithDataFieldValue');
      } else {
        copyWithConstructorLogicSet.add('$fieldName: $copyWithDataFieldValue,');
      }
    }

    superTypeVisitor(classElement.supertype, fieldAnalyzer);
    classElement.fields.forEach(fieldAnalyzer);

    StringBuffer copyWithFunctionParams = StringBuffer('');
    for (final item in copyWithFunctionParamsSet) {
      copyWithFunctionParams.writeln(item);
    }

    StringBuffer copyWithConstructorLogic = StringBuffer('');
    for (final item in copyWithConstructorLogicSet) {
      copyWithConstructorLogic.writeln(item);
    }

    StringBuffer copyWithExtraLogic = StringBuffer('');
    for (final item in copyWithExtraLogicSet) {
      copyWithExtraLogic.writeln(item);
    }

    return '''
      /// This Extension on [$classType] is to generate the code for a copyWith(...) function.
      extension \$${classType}CopyWithExtension on $classType {
          $classType copyWith({
          ${copyWithFunctionParams.toString()}
        }) {
          return $classType(
            ${copyWithConstructorLogic.toString()}
          )${copyWithExtraLogic.isNotEmpty ? copyWithExtraLogic.toString() : ''};
        }
      }
    ''';
  }

  /// Gets the CopyWith annotation for a given ClassElement
  CopyWith? _getCopyWith(ClassElement classElement) {
    DartObject? copyWithAnnotation = annotationChecker
        .firstAnnotationOf(classElement, throwOnUnresolved: false);

    if (copyWithAnnotation != null) {
      return CopyWith(
        ignoreSynthetic:
            copyWithAnnotation.getField('ignoreSynthetic')?.toBoolValue(),
        allowNullable:
            copyWithAnnotation.getField('allowNullable')?.toBoolValue(),
        avoidConstructor:
            copyWithAnnotation.getField('avoidConstructor')?.toBoolValue(),
        considerEmptyCollectionAsNull: copyWithAnnotation
            .getField('considerEmptyCollectionAsNull')
            ?.toBoolValue(),
        includePrivateFields:
            copyWithAnnotation.getField('includePrivateFields')?.toBoolValue(),
      );
    }
    return null;
  }

  /// Gets the CopyWithKey annotation for a given FieldElement
  CopyWithKeyConfig? _getCopyWithKey(FieldElement field) {
    Iterable<DartObject> annotations =
        keyAnnotationChecker.annotationsOf(field, throwOnUnresolved: false);
    int annotationsCount = annotations.length;
    if (annotationsCount > 1) {
      throw InvalidGenerationSourceError(
          'Repeated `@CopyWithKey` annotation on ${field.displayName}.',
          todo: 'Check the `@CopyWithKey` annotations on field '
              '${field.displayName}.',
          element: field);
    }
    if (annotationsCount == 1) {
      DartObject annotation = annotations.first;

      String? copyWithFunction;
      final function =
          annotation.getField('copyWithFunction')?.toFunctionValue();
      if (function != null) {
        String enclosingClass = function.enclosingElement.displayName;
        String functionName = function.displayName;
        String functionParams =
            'currentData: ${!field.isPrivate ? 'this.' : ''}${field.displayName}, newData: ${field.displayName.substring(field.isPrivate ? 1 : 0)}';
        copyWithFunction = '$enclosingClass.$functionName($functionParams)';
      }

      Object? defaultValue = annotationValue(annotation, field, 'defaultValue');
      final key = CopyWithKeyConfig(
        ignoreSynthetic: annotation.getField('ignoreSynthetic')?.toBoolValue(),
        ignore: annotation.getField('ignore')?.toBoolValue(),
        includeIfPrivate:
            annotation.getField('includeIfPrivate')?.toBoolValue(),
        avoidConstructor:
            annotation.getField('avoidConstructor')?.toBoolValue(),
        allowNullable: annotation.getField('allowNullable')?.toBoolValue(),
        considerEmptyCollectionAsNull:
            annotation.getField('considerEmptyCollectionAsNull')?.toBoolValue(),
        defaultValue: defaultValue,
        copyWithFunction: copyWithFunction,
      );
      return key;
    }
    return null;
  }
}
