import 'dart:async';

import 'package:adeptannotations/adeptannotations.dart';
import 'package:adeptgenerator/src/base_generator_for_annotation.dart';
import 'package:adeptgenerator/src/config/is_empty_key_config.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

class IsEmptyGenerator extends BaseGeneratorForAnnotation<IsEmpty, IsEmptyKey> {
  final IsEmpty globalAnnotation;
  final bool defaultIgnoreSynthetic = false;
  final bool defaultIncludeIfPrivate = false;

  IsEmptyGenerator(BuilderOptions options)
      : globalAnnotation = IsEmpty(
          ignoreSynthetic: options.config['is_empty_ignore_synthetic'],
          includePrivateFields:
              options.config['is_empty_include_private_fields'],
        );

  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement =
        getClassElement(element, annotationName: '@IsEmpty()');
    String isEmptyFunction = _generateIsEmptyFunction(classElement, annotation);
    return '''
    $isEmptyFunction
    ''';
  }

  /// This function manages the logic of generating the isEmpty boilerplate code
  String _generateIsEmptyFunction(
      ClassElement classElement, ConstantReader annotation) {
    String classType = classElement.displayName;
    String functionName = '\$${classType}IsEmpty';

    Set<String> isEmptyFunctionLogicSet = {};
    final isEmptyAnnotation = _getIsEmpty(classElement) ?? const IsEmpty();

    void fieldAnalyzer(FieldElement field) {
      final isEmptyKey = _getIsEmptyKey(field) ?? const IsEmptyKeyConfig();
      final ignoreSynthetic = isEmptyKey.ignoreSynthetic ??
          isEmptyAnnotation.ignoreSynthetic ??
          globalAnnotation.ignoreSynthetic ??
          defaultIgnoreSynthetic;
      final includeIfPrivate = isEmptyKey.includeIfPrivate ??
          isEmptyAnnotation.includePrivateFields ??
          globalAnnotation.includePrivateFields ??
          defaultIncludeIfPrivate;

      if (field.isSynthetic && ignoreSynthetic) return;
      if (field.isConst || field.isStatic) return;
      if (field.isPrivate && !includeIfPrivate) return;
      if (isEmptyKey.ignore) return;

      final isNullable = field.type.nullabilitySuffix != NullabilitySuffix.none;
      final isCollection = field.type.isDartCoreList ||
          field.type.isDartCoreSet ||
          field.type.isDartCoreMap ||
          field.type.isDartCoreIterable;

      String fieldName = 'instance.${field.displayName}';
      String validation = '';

      if (isEmptyKey.isEmptyFunction != null) {
        validation = isEmptyKey.isEmptyFunction!;
      } else if (isCollection) {
        if (isNullable) {
          validation = '($fieldName?.isEmpty ?? true)';
        } else {
          validation = '$fieldName.isEmpty';
        }
      } else if (isNullable) {
        validation = '$fieldName == null';
      } else {
        return;
      }

      isEmptyFunctionLogicSet.add(validation);
    }

    superTypeVisitor(classElement.supertype, fieldAnalyzer);
    classElement.fields.forEach(fieldAnalyzer);

    StringBuffer isEmptyFunctionLogic = StringBuffer('');
    int i = isEmptyFunctionLogicSet.length;
    for (final item in isEmptyFunctionLogicSet) {
      isEmptyFunctionLogic.writeln(item);
      if (--i > 0) {
        isEmptyFunctionLogic.write(' &&');
      } else {
        isEmptyFunctionLogic.write(';');
      }
    }

    return '''
      /// This function is to generate the isEmpty boilerplate code.
      bool $functionName($classType instance) => ${isEmptyFunctionLogic.toString()}
    
      /// This Extension on [$classType] is to generate the isEmpty and isNotEmpty functions.
      extension ${functionName}Extension on $classType {
          bool get isEmpty => $functionName(this);
          bool get isNotEmpty => !isEmpty;
      }
    ''';
  }

  /// Gets the IsEmpty annotation for a given ClassElement
  IsEmpty? _getIsEmpty(ClassElement classElement) {
    DartObject? isEmptyAnnotation = annotationChecker
        .firstAnnotationOf(classElement, throwOnUnresolved: false);

    if (isEmptyAnnotation != null) {
      return IsEmpty(
        ignoreSynthetic:
            isEmptyAnnotation.getField('ignoreSynthetic')?.toBoolValue(),
        includePrivateFields:
            isEmptyAnnotation.getField('includePrivateFields')?.toBoolValue(),
      );
    }
    return null;
  }

  /// Gets the IsEmptyKey annotation for a given FieldElement
  IsEmptyKeyConfig? _getIsEmptyKey(FieldElement field) {
    Iterable<DartObject> annotations =
        keyAnnotationChecker.annotationsOf(field, throwOnUnresolved: false);
    int annotationsCount = annotations.length;
    if (annotationsCount > 1) {
      throw InvalidGenerationSourceError(
          'Repeated `@IsEmptyKey` annotation on ${field.displayName}.',
          todo: 'Check the `@IsEmptyKey` annotations on field '
              '${field.displayName}.',
          element: field);
    }
    if (annotationsCount == 1) {
      DartObject annotation = annotations.first;

      String? isEmptyFunction;
      final function =
          annotation.getField('isEmptyFunction')?.toFunctionValue();
      if (function != null) {
        String enclosingClass = function.enclosingElement.displayName;
        String functionName = function.displayName;
        String functionParams = 'field: instance.${field.displayName}';
        isEmptyFunction = '$enclosingClass.$functionName($functionParams)';
      }

      final key = IsEmptyKeyConfig(
        ignoreSynthetic: annotation.getField('ignoreSynthetic')?.toBoolValue(),
        ignore: annotation.getField('ignore')?.toBoolValue(),
        includeIfPrivate:
            annotation.getField('includeIfPrivate')?.toBoolValue(),
        isEmptyFunction: isEmptyFunction,
      );
      return key;
    }
    return null;
  }
}
